;;; C bindings for Scheme->C

(eval-when 
 (compile)

 (define renaming-rules '())

 (define gensym-data 0)
 (define (gensym x)
   (set! gensym-data (+ gensym-data 1))
   (string->symbol (format "gensym-~a-~s-~a" module-name gensym-data x)))
 
 (define (zip a b)
   (if (or (null? a) (null? b))
       '()
       (cons (list (car a) (car b)) (zip (cdr a) (cdr b)))))
 
 (define (c-ify x)
   (list->string (map (lambda (x) (if (eq? x #\-) #\_ x)) (string->list x))))

 ;; Lifted from QobiScheme 
 (define (map-n f n)
   ;; needs work: To eliminate REVERSE.
   (let loop ((i 0) (c '()))
     (if (< i n) (loop (+ i 1) (cons (f i) c)) (reverse c))))

 (define (c-out-convert type) 
   (case type
     ((int) 'S2CINT_TSCP)
     ((char) 'CHAR_TSCP)
     ((uint) 'S2CUINT_TSCP)
     ((double) 'DOUBLE_TSCP)
     ((pointer) 'POINTER_TSCP)
     ((string) 'CSTRING_TSCP)
     (else (error (format "Can't convert ~a to Scheme from C" type)))))
 (define (c-in-convert x) 
   (case x
     ((int) 'TSCP_S2CINT)
     ((char) 'TSCP_CHAR)
     ((uint) 'TSCP_S2CUINT)
     ((double) 'TSCP_DOUBLE)
     ((pointer) 'TSCP_POINTER)
     ((string) 'TSCP_POINTER)
     (else (error (format "Can't convert ~a to C from Scheme" x)))))

 (define (append-symbols . s)
   (string->symbol (reduce string-append  (map symbol->string s) "")))

 (define (c-set-convert x) 
   (case x
     ((int) 'c-int-set!)
     ((pointer) 'c-s2cuint-set!)
     (else (error 'c-set-convert "doesn't support ~a" x))))
 (define (c-ref-convert x) 
   (case x
     ((int) 'c-int-ref)
     ((pointer) 'c-s2cuint-ref)
     (else (error 'c-set-convert "doesn't support ~a" x))))

 (define (schemeify-name x)
   (if (assoc x renaming-rules)
       (second (assoc x renaming-rules))
       (list->string (map (lambda (x) (if (eq? x #\_)
					  #\-
					  x)) (string->list x)))))

 (define (string->pretty-symbol s)
   (string->symbol 
    (list->string 
     (map char-upcase (string->list s))))))

;;; horribly ugly hack to get an #include at the toplevel
;;; only use at the toplevel, otherwise unspeakable horrors will unfold
(define-macro c-include
  (lambda (f e)
    `(define (,(gensym 0)) ((lap () ,(string-append "0 );} 
  #include<" (string-append 
	      (second f) 
	      (let ((dummy (c-ify (symbol->string (gensym 0)))))
		(string-append "> 
int " (string-append dummy 
		     (string-append 
		      "(){ return 1;" 
		      (string-append
		       dummy "("))))))))))))

;;; Required for offsetof
(c-include "stddef.h")

(define-macro struct-offset
  (lambda (f e)
    (e `((lap () (C_FIXED ("offsetof" 
			   ,(string-append "struct " (second f)) 
			   ,(third f))))) e)))

(define-macro c-value
  (lambda (f e) `((lap () (,(c-out-convert (second f)) ,(third f))))))
(define-macro c-function
  (lambda (f e)
    (e (let* ((c-vars (cdaddr f))
	      (vars (map-n gensym (length c-vars)))
	      (s-vars (map-n gensym (length c-vars))))
	 `(lambda ,s-vars 
	    ,(let 
		 ((body 
		   `((lap ,vars
			  ,((if (eq? (second f) 'void)
				(lambda (x) x)
				(lambda (x) `(,(c-out-convert (second f)) ,x)))
			    `(,(caaddr f) 
			      ,@(map (lambda (x) (list (c-in-convert (car x)) 
						       (second x)))
				     (zip c-vars vars))))) ,@s-vars)))
	       (if (eq? (second f) 'void)
		   `(begin ,body ((lap () FALSEVALUE)))
		   body)))) e)))

(define-macro c-define-field
  (lambda (f e)
    (e (let ((struct (second f))
	     (field (third f))
	     (field-type (fourth f))
	     (offst (fifth f)))
	 `(begin
	    (define (,(append-symbols struct '- field '- 'set!) o v)
	      (,(c-set-convert field-type) o ,offst v))
	    (define (,(append-symbols struct '- field) o)
	      (,(c-ref-convert field-type) o ,offst))
	    (define (,(append-symbols struct '- field '- 'update) o f)
	      (,(append-symbols struct '- field '- 'set!) o
	       (f (,(append-symbols struct '- field) o)))))) e)))

(define-macro c-define-struct-field-full
  (lambda (f e)
    (e (let* ((struct      (second f))
	      (struct-name (third f))
	      (field       (fourth f))
	      (field-name  (fifth f))
	      (field-type  (sixth f))
	      (offst `(struct-offset ,struct-name ,field-name)))
	 `(begin
	    (define (,(append-symbols struct '- field '- 'set!) o v)
	      (,(c-set-convert field-type) o ,offst v))
	    (define (,(append-symbols struct '- field) o)
	      (,(c-ref-convert field-type) o ,offst))
	    (define (,(append-symbols struct '- field '- 'update) o f)
	      (,(append-symbols struct '- field '- 'set!) o
	       (f (,(append-symbols struct '- field) o)))))) e)))

(define-macro c-define-struct-field
  (lambda (f e)
    (e (let* ((struct-name (second f))
	      (struct (string->pretty-symbol (schemeify-name struct-name)))
	      (field-name (third f))
	      (field (string->pretty-symbol (schemeify-name field-name)))
	      (type (fourth f)))
	 `(c-define-struct-field-full ,struct ,struct-name 
				      ,field ,field-name ,type)) e)))

(define-macro c-sizeof
  (lambda (f e)
    (e `((lap () (C_FIXED ("sizeof" ,(second f))))) e)))

(define-macro add-id-rename-rules
  (lambda (f e)
    (e `(eval-when (compile)
		   (set! renaming-rules
			 (append renaming-rules ,(second f)))) e)))


;;; Get rid of a malloc warning
(c-include "stdlib.h")
(c-include "string.h")

